#include "PrecompiledHeaders.h"

#include "Hair.h"

#define SIMD_PI_2 SIMD_PI*0.5
float Hair::segmentMass = 5.f;

Hair::Hair(const btVector3 &origin, const btQuaternion &orientation, GlobalPhysics *physicsEngine, float inRadius, float inHeight)
	: m_origin(origin), m_orientation(orientation), m_physicsEngine(physicsEngine), radius(inRadius), height(inHeight) {
	
	// create collision shape
	btScalar totalHeight = 2 * radius + height;
	m_collisionShape = new btCapsuleShape(radius,height);
	btVector3 localInertia;
	m_collisionShape->calculateLocalInertia(segmentMass, localInertia);

	btTransform hairFrame;
	hairFrame.setOrigin(m_origin);
	hairFrame.setRotation(m_orientation);

	btTransform hairSegmentOffset;
	hairSegmentOffset.setIdentity();
	hairSegmentOffset.setOrigin(btVector3(0,totalHeight,0));
	btTransform currentHairSegmentFrame = hairFrame;

	for(int i = 0; i < SEGMENT_COUNT; i++) {
		btRigidBody::btRigidBodyConstructionInfo segmentCI(i==0 ? 0.f : segmentMass,new btDefaultMotionState(currentHairSegmentFrame),m_collisionShape,localInertia);
		m_segments.push_back(new btRigidBody(segmentCI));
		float ratio = float(SEGMENT_COUNT-i)/float(SEGMENT_COUNT+1);
		m_segments[i]->setDamping(0.8f+1.5f*ratio,0.3f);
		physicsEngine->addRigidBody(m_segments[i]);
		currentHairSegmentFrame = currentHairSegmentFrame * hairSegmentOffset;
	}

	// add constraints to join nodes
	btScalar constraintOffset = 1;
	btTransform frameInA, frameInB;
	frameInA.setIdentity();
	frameInB.setIdentity();
	frameInA.setOrigin(btVector3(0,(totalHeight/2.0f+constraintOffset),0));
	frameInB.setOrigin(btVector3(0,-(totalHeight/2.0f+constraintOffset),0));
	for(int i = 0; i < CONSTRAINT_COUNT; i++) {
		m_constraints.push_back(new btConeTwistConstraint(*m_segments[i], *m_segments[i+1], frameInA, frameInB));
		float ratio = float(CONSTRAINT_COUNT-i)/float(CONSTRAINT_COUNT+1);
		float limitRatio = (1.0f-ratio+0.1f)*0.6f;
		m_constraints[i]->setLimit(btScalar(SIMD_PI_2*limitRatio), btScalar(SIMD_PI_2*limitRatio), 0.0);
		physicsEngine->addConstraint(m_constraints[i]);
	}
}

Hair::Hair(Hair&& other) : m_origin(other.m_origin), m_orientation(other.m_orientation), 
	m_physicsEngine(other.m_physicsEngine), m_collisionShape(other.m_collisionShape),
	radius(other.radius), height(other.height) {
		m_segments = std::move(other.m_segments);
		m_constraints = std::move(other.m_constraints);
        other.m_physicsEngine = nullptr;
		other.m_collisionShape = nullptr;
    }

Hair::Hair(const Hair& other) : m_physicsEngine(other.m_physicsEngine), m_origin(other.m_origin), m_orientation(other.m_orientation), radius(other.radius), height(other.height) {
	throw std::logic_error("Copy constructor not implemented.\n Should never enter here.\n");
}

Hair::~Hair() {
	for(unsigned int i = 0; i < m_constraints.size(); i++) {
		m_physicsEngine->removeConstraint(m_constraints[i]);
		delete m_constraints[i];
	}
	for(unsigned int i = 0; i < m_segments.size(); i++) {
		m_physicsEngine->removeRigidBody(m_segments[i]);
		delete m_segments[i]->getMotionState();
		delete m_segments[i];
	}
	if(m_collisionShape != nullptr) {
		delete m_collisionShape;
	}
}
