#include "PrecompiledHeaders.h"

#include "CallbackMenu.h"
#include "SeaAnemone.h"
#include "TextureGenerator.h"
#include "Resources.h"

using namespace ci;
using namespace ci::app;
using namespace std;

#define LIGHTS_COUNT 3

class SeaAnemonesApp : public AppNative {
public:
    void prepareSettings(Settings *settings);
    void setup();

    void resize();

    void mouseMove(MouseEvent event);
    void mouseDown(MouseEvent event);
    void mouseDrag(MouseEvent event);
    void mouseUp(MouseEvent event);
    void keyDown(KeyEvent event);

    void addColor();
    void removeColor();
    void createMenu();

    void update();
    void draw();

    void render();
    void renderAnemone(Vec3f camera);
    void renderAnemoneGradient();
    void renderAnemoneBump();

    void createDepthMap();

    void loadTextures(string folderPath, vector<gl::Texture>& textures, vector<string>& textureNames, GLenum wrap=GL_REPEAT);

    MayaCamUI mCamera;
    CameraPersp	mRenderCam;
    float farPlane;
    Vec2i mMousePos;

    params::CallbackMenuRef	mParams;
    Font mFont;

    float mTime, mPrevTime;
    float mAnimTime;

    bool mAnimate;

    gl::GlslProg mPhongShader;
    gl::Texture mTexture;

    gl::Material mMaterial;
    gl::Material mGroundMaterial;

    GLint attr_height;
    gl::Texture mGradient;

    gl::Fbo::Format mFormat;
    gl::Fbo mGradientFbo;

    float mTextureAlpha;

    static const int mTextureSize;

    //debug only
    bool mViewTexture;

    SeaAnemone mSeaAnemone;

    gl::Fbo	mFboScene;
    gl::Fbo	mFboBlur1;
    gl::Fbo	mFboBlur2;
    gl::GlslProg mShaderBlur;
    bool mDrawBloom;
	bool mDrawCapsules;
    static const int mFboSize;

    //bump mapping
    gl::GlslProg mBumpGradientShader;
    gl::GlslProg mBumpShader;
    gl::Texture mNormalMapTex; //currently using

    vector<gl::Texture> mNormalTextures;
    vector<gl::Texture> mNoiseTextures;
    vector<string> mNormalTextureNames;
    vector<string> mNoiseTextureNames;

    int mNormalMapCurrent; //currentyly using that index
    int mNormalMapChosen; //user chose that from menu

    bool mBumpMapOn;

    //sss depth map = SSAO
    gl::GlslProg mNormalDepthShader;
    gl::Fbo mFboSceneNormalDepth;
    bool mViewDepthTexture;
    bool mViewNormalDepth;
    gl::Fbo::Format mSSAOFormat;

    float mFinalStrength;
    float mTotStrength;
    float mStrength;
    float mOffset;
    float mFalloff;
    float mRad;

    //sss
    gl::Fbo mFboSceneAO;
    gl::GlslProg mSssShader;
    gl::GlslProg mSSAOShader;
    bool mSssOn;

    float mPower;
    float mScale;
    float mDisortion;
    float mAmbient;
    float mShininess;
    Color mDiffuse;
    float mKC;
    float mKL;
    float mKQ;
    float mSpecScale;

    Vec3f mLightPositions[LIGHTS_COUNT];
    Vec3f mLightColors[LIGHTS_COUNT]; //vec not color because you can't pass an array of colors to shader in cinder
};

const int SeaAnemonesApp::mTextureSize = 256;
const int SeaAnemonesApp::mFboSize = 64;

void SeaAnemonesApp::prepareSettings(Settings *settings)
{
    settings->setWindowSize(875, 700);
    settings->setFrameRate(60.0f);
    settings->setBorderless(true);
}

void SeaAnemonesApp::addColor()
{
    int last = mSeaAnemone.colors.size();
    for(int i = 0; i < last; ++i){
        stringstream label;
        label << "Color" << i;
        mParams->removeParam(label.str());
        label << " starts at";
        mParams->removeParam(label.str());
    }

    mSeaAnemone.colors.push_back(Vec3f(0, 0, 0));
    mSeaAnemone.colorStartsAt.push_back(1.0f);

    for(int i = 0; i <= last; ++i){
        stringstream label;
        label << "Color" << i;
        mParams->addParam(label.str(), (Color*)&mSeaAnemone.colors[i]);
        label << " starts at";
        mParams->addParam(label.str(), &(mSeaAnemone.colorStartsAt[i]), "min=0 max=1.0 step=0.05");
    }
}

void SeaAnemonesApp::removeColor()
{
    int last = mSeaAnemone.colors.size();
    if(last == 2) return;

    stringstream label;
    label << "Color" << last-1;
    mParams->removeParam(label.str());
    label << " starts at";
    mParams->removeParam(label.str());

    mSeaAnemone.colors.pop_back();
    mSeaAnemone.colorStartsAt.pop_back();
}

void SeaAnemonesApp::loadTextures(string folderPath, vector<gl::Texture>& textures, vector<string>& textureNames, GLenum wrap){
    textures.clear();
    textureNames.clear();

    fs::path p(folderPath);
    for(fs::directory_iterator it(p); it!=fs::directory_iterator(); ++it)
    {
        if (fs::is_regular_file(*it))
        {
            string fileName =  it->path().filename().string();
            textures.push_back(gl::Texture(loadImage(folderPath + fileName)));
            textures[textures.size()-1].setWrap(wrap, wrap);
            textures[textures.size()-1].setMinFilter(GL_LINEAR);
            textures[textures.size()-1].setMagFilter(GL_LINEAR);

            fileName = fileName.substr(0, fileName.rfind(".")); //without extension
            textureNames.push_back(fileName);
        }
    }
}

void SeaAnemonesApp::createMenu(){
    mParams = params::CallbackMenu::create("Sea Anemone", Vec2i(200, 650));
    mParams->addSeparator();
    mParams->addParam("FPS",
        std::function<void(float)>([this](float x) {}),
        std::function<float()>([this]() -> float {
            return this->getAverageFps();}));
    mParams->addSeparator();
    mParams->addParam("r",
        std::function<void(float)>([this](float x) {
            this->mSeaAnemone.r = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<float()>([this]() -> float {
            return this->mSeaAnemone.r;}),
        "min=0 step=5");
    mParams->addParam("R",
        std::function<void(float)>([this](float x) {
            this->mSeaAnemone.R = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<float()>([this]() -> float {
            return this->mSeaAnemone.R;}),
        "min=0.1 step=5");
    mParams->addParam("H",
        std::function<void(float)>([this](float x) {
            this->mSeaAnemone.H = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<float()>([this]() -> float {
            return this->mSeaAnemone.H;}),
        "min=0.1 step=10");
    mParams->addSeparator();
    mParams->addParam("middle R",
        std::function<void(float)>([this](float x) {
            this->mSeaAnemone.mr = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<float()>([this]() -> float {
            return this->mSeaAnemone.mr;}),
        "min=0.1 step=5");
    mParams->addParam("middle R H",
        std::function<void(float)>([this](float x) {
            this->mSeaAnemone.mrH = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<float()>([this]() -> float {
            return this->mSeaAnemone.mrH;}),
        "min=0.0 max=1.0 step=0.1");
    mParams->addParam("vertical scale",
        std::function<void(float)>([this](float x) {
            this->mSeaAnemone.verticalScale = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<float()>([this]() -> float {
            return this->mSeaAnemone.verticalScale;}),
        "min=0.0 max=1.0 step=0.05");
    mParams->addSeparator();
    mParams->addParam("Radial segments",
        std::function<void(int)>([this](int x) {
            this->mSeaAnemone.radialSegments = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<int()>([this]() -> int {
            return this->mSeaAnemone.radialSegments;}),
        "min=1 step=10");
    mParams->addParam("Height segments",
        std::function<void(int)>([this](int x) {
            this->mSeaAnemone.heightSegments = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<int()>([this]() -> int {
            return this->mSeaAnemone.heightSegments;}),
        "min=1 step=10");
    mParams->addSeparator();
    mParams->addParam("sole clamp",
        std::function<void(float)>([this](float x) {
            this->mSeaAnemone.soleClamp = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<float()>([this]() -> float {
            return this->mSeaAnemone.soleClamp;}),
        "min=0.0 max=1.0 step=0.1");
    mParams->addParam("wave size",
        std::function<void(float)>([this](float x) {
            this->mSeaAnemone.waveD = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<float()>([this]() -> float {
            return this->mSeaAnemone.waveD;}),
        "min=0 step=0.02");
    mParams->addParam("num waves",
        std::function<void(int)>([this](int x) {
            this->mSeaAnemone.numWaves = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<int()>([this]() -> int {
            return this->mSeaAnemone.numWaves;}),
        "min=0 max=20 step=1");
    mParams->addSeparator();
    mParams->addParam("Bend Angle",
        std::function<void(float)>([this](float x) {
            this->mSeaAnemone.bendAngleDeg = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<float()>([this]() -> float {
            return this->mSeaAnemone.bendAngleDeg;}),
        "min=0. max=180. step=1.");
    mParams->addParam("Bend Axis",
        std::function<void(Vec3f)>([this](Vec3f x) {
            this->mSeaAnemone.bendAxis = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<Vec3f()>([this]() -> Vec3f {
            return this->mSeaAnemone.bendAxis;}),
        "opened=1");
    mParams->addSeparator();
    mParams->addParam("Hair in line",
        std::function<void(int)>([this](int x) {
            this->mSeaAnemone.hairInLine = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<int()>([this]() -> int {
            return this->mSeaAnemone.hairInLine;}),
        "min=0 step=1");
    mParams->addParam("Number of lines",
        std::function<void(int)>([this](int x) {
            this->mSeaAnemone.hairLines = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<int()>([this]() -> int {
            return this->mSeaAnemone.hairLines;}),
        "min=0 max=20 step=1");
    mParams->addParam("hair R",
        std::function<void(float)>([this](float x) {
            this->mSeaAnemone.hairR = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<float()>([this]() -> float {
            return this->mSeaAnemone.hairR;}),
        "min=0.1 step=1");
    mParams->addParam("hair H",
        std::function<void(float)>([this](float x) {
            this->mSeaAnemone.hairH = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<float()>([this]() -> float {
            return this->mSeaAnemone.hairH;}),
        "min=0.1 step=1");
    mParams->addParam("hair cylindrical",
        std::function<void(bool)>([this](bool x) {
            this->mSeaAnemone.hairCylindrical = x;
            this->mSeaAnemone.regenerate((float)this->getElapsedSeconds());}),
        std::function<bool()>([this]() -> bool {
            return this->mSeaAnemone.hairCylindrical;}));
    mParams->addSeparator();
    mParams->addParam("texture alpha", &(mTextureAlpha), "min=0.01 max=1.0 step=0.05");
    mParams->addParam("view texture", &(mViewTexture));
    mParams->addParam("bump mapping", &(mBumpMapOn));
    mParams->addParam("normal map", mNormalTextureNames, &mNormalMapChosen);
    mParams->addSeparator();
    mParams->addParam("view depth texture", &(mViewDepthTexture));
    mParams->addParam("view normal depth", &(mViewNormalDepth));
    mParams->addParam("SSAO darkness", &(mFinalStrength), "min=0.0 max=1.0 step=0.1");
    mParams->addParam("SSAO totStrength", &(mTotStrength), "min=0.0 max=1.0 step=0.01");
    mParams->addParam("SSAO strength", &(mStrength), "min=0.0 max=1.0 step=0.1");
    mParams->addParam("SSAO offset", &(mOffset), "min=0.0 max=1.0 step=0.001");
    mParams->addParam("SSAO falloff", &(mFalloff), "min=0.0 max=1.0 step=0.1");
    mParams->addParam("SSAO rad", &(mRad), "min=0.0 max=1.0 step=0.01");
    mParams->addSeparator();
    mParams->addParam("use SSS", &(mSssOn));
    mParams->addParam("Power", &(mPower), "min=0.0 step=1.0");
    mParams->addParam("Scale", &(mScale), "min=0.0 step=1.0");
    mParams->addParam("Disortion", &(mDisortion), "min=0.0 max=1.0 step=0.005");
    mParams->addParam("Ambient", &(mAmbient), "min=0.0 step=0.5");
    mParams->addParam("Shininess", &(mShininess), "min=0.0 step=10.0");
    mParams->addParam("Diffuse", &(mDiffuse));
    mParams->addParam("KC", &(mKC), "min=0.0 max=1.0 step=0.1");
    mParams->addParam("KL", &(mKL), "min=0.0 max=1.0 step=0.05");
    mParams->addParam("KQ", &(mKQ), "min=0.0 max=1.0 step=0.0001");
    mParams->addParam("Spec scale", &(mSpecScale), "min=0.0 max=30.0 step=0.1");
    mParams->addSeparator();
    mParams->addParam("view triangles", &(mSeaAnemone.viewTriangles));
    mParams->addParam("view normals", &(mSeaAnemone.viewNormals));
    mParams->addParam("view model", &(mSeaAnemone.viewModel));
	mParams->addParam("view hair", &(mSeaAnemone.viewHair));
	mParams->addParam("view capsules", &(mDrawCapsules));
    mParams->addParam("view bloom", &(mDrawBloom));
    mParams->addParam("animation", &(mAnimate));
    mParams->addParam("normals lengths", &(mSeaAnemone.normalLength), "min=1 step=2");
    mParams->addSeparator();
    mParams->addButton("Add color", std::bind(&SeaAnemonesApp::addColor, this));
    mParams->addButton("Remove color", std::bind(&SeaAnemonesApp::removeColor, this));
    for(unsigned int i = 0; i < mSeaAnemone.colors.size(); ++i){
        stringstream label;
        label << "Color" << i;
        mParams->addParam(label.str(), (Color*)&mSeaAnemone.colors[i]);
        label << " starts at";
        mParams->addParam(label.str(), &(mSeaAnemone.colorStartsAt[i]), "min=0 max=1.0 step=0.05");
    }
}

void SeaAnemonesApp::setup()
{
    glEnable(GL_RESCALE_NORMAL);

    farPlane = 1000.0f;
    mRenderCam = CameraPersp(getWindowWidth(), getWindowHeight(), 60.0f, 10.0f, farPlane);
    mRenderCam.setEyePoint(Vec3f(0.0f, 100.0f, 600.0f));
    mRenderCam.setCenterOfInterestPoint(Vec3f(0.0f, 100.0f, 0.0f));
    mCamera.setCurrentCam(mRenderCam);

    //setup material - change later for NICE
    mMaterial.setAmbient(Color(0.0f, 0.05f, 0.05f));
    mMaterial.setDiffuse(Color(0.4f, 0.5f, 0.5f));
    mMaterial.setSpecular(Color::black());
    mMaterial.setShininess(0.0f);

    mGroundMaterial.setAmbient(Color(0.2125f, 0.1275f, 0.054f));
    mGroundMaterial.setDiffuse(Color(0.714f, 0.4284f, 0.18144f));
    mGroundMaterial.setSpecular(Color(0.3935f, 0.2719f, 0.1667f));
    mGroundMaterial.setShininess(25.6f);

    mFboScene = gl::Fbo(getWindowWidth(), getWindowHeight());
    mFboBlur1 = gl::Fbo(mFboSize, mFboSize);
    mFboBlur2 = gl::Fbo(mFboSize, mFboSize);
    mFboSceneNormalDepth = gl::Fbo((int)(getWindowWidth()/2.0f), (int)(getWindowHeight()/2.0f));
    mFboSceneAO = gl::Fbo((int)(getWindowWidth()/2.0f), (int)(getWindowHeight()/2.0f));
    mGradientFbo = gl::Fbo(getWindowWidth(), 1, mFormat);

    //load and compile shader
    try {
        mShaderBlur = gl::GlslProg(loadAsset("shaders/blur_vert.glsl"), loadAsset("shaders/blur_frag.glsl"));
        mPhongShader = gl::GlslProg(loadAsset("shaders/phong_vert.glsl"), loadAsset("shaders/phong_frag.glsl"));
        mBumpGradientShader = gl::GlslProg(loadAsset("shaders/bumpGrad_vert.glsl"), loadAsset("shaders/bumpGrad_frag.glsl"));
        mBumpShader = gl::GlslProg(loadAsset("shaders/bump_vert.glsl"), loadAsset("shaders/bump_frag.glsl"));
        mNormalDepthShader = gl::GlslProg(loadAsset("shaders/normaldepth_vert.glsl"), loadAsset("shaders/normaldepth_frag.glsl"));
        mSSAOShader = gl::GlslProg(loadAsset("shaders/ssao_vert.glsl"), loadAsset("shaders/ssao_frag.glsl"));
        mSssShader = gl::GlslProg(loadAsset("shaders/sss_vert.glsl"), loadAsset("shaders/sss_frag.glsl"));
        attr_height = mPhongShader.getAttribLocation("in_height");

        loadTextures(getAssetPath("normalMaps/").string(), mNormalTextures, mNormalTextureNames);
        loadTextures(getAssetPath("noise/").string(), mNoiseTextures, mNoiseTextureNames);
    }
    catch( const std::exception &e ) {
        app::console() << "Could not load and compile resource:" << e.what() << std::endl;
        return;
    }

    mNormalMapTex = mNormalTextures[0];
    mNormalMapCurrent = 0;
    mNormalMapChosen = 0;

    //menu
    createMenu();

    mFont = Font( "Quicksand Book Regular", 12.0f );

    //time
    mTime = mPrevTime = 0.0f;
    mAnimTime = 0.0f;

    mAnimate = true;
	mSeaAnemone.viewModel = true;
    mSeaAnemone.regenerate(0.f);

    mTextureAlpha = 0.8f;
    mViewTexture = false;

    mDrawBloom = false;
	mDrawCapsules = false;
    mBumpMapOn = true;
    mViewDepthTexture = false;
    mViewNormalDepth = false;
    mSssOn = true;

    mFinalStrength = 0.7f;
    mTotStrength = 0.38f;
    mStrength = 0.3f;
    mOffset = 0.002f;
    mFalloff = 0.0f;
    mRad = 0.03f;

    mPower = 4.0f;
    mScale = 9.0f;
    mDisortion = 0.185f;
    mAmbient = 2.0f;
    mShininess = 0.0f;
    mDiffuse = Color(0.85f, 0.63f, 0.51f);
    mKC = 0.5f;
    mKL = 0.05f;
    mKQ = 0.0001f;
    mSpecScale = 0.0f;

    TextureGenerator::perlinNoiseTexture(mTexture, mTextureSize);
    mTexture.setWrap(GL_REPEAT, GL_REPEAT);

    mSSAOFormat.setColorInternalFormat(GL_RGBA16F_ARB);
    mSSAOFormat.setSamples(4); // enable 4x antialiasing

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
}

void SeaAnemonesApp::resize()
{
    //setup the camera
    CameraPersp cam = mCamera.getCamera();
    cam.setPerspective(60.0f, getWindowAspectRatio(), 10.f, 1000.0f);
    mCamera.setCurrentCam(cam);

    mFboScene = gl::Fbo(getWindowWidth(), getWindowHeight());
    mFboSceneNormalDepth = gl::Fbo(getWindowWidth(), getWindowHeight(), mSSAOFormat);
    mFboSceneAO = gl::Fbo(getWindowWidth(), getWindowHeight(), mSSAOFormat);
    mGradientFbo = gl::Fbo(getWindowWidth(), 1, mFormat);
}

void SeaAnemonesApp::keyDown(KeyEvent event)
{
    switch(event.getCode()) {
    case KeyEvent::KEY_ESCAPE:
        quit();
        break;
    case KeyEvent::KEY_f:
        setFullScreen(!isFullScreen());
        break;
    case KeyEvent::KEY_u:
        mSeaAnemone.update(mAnimTime);
        break;
	case KeyEvent::KEY_SPACE:
		mSeaAnemone.m_physics.ballReset();
		break;
    }
}

void SeaAnemonesApp::mouseMove(MouseEvent event)
{
    mMousePos = event.getPos();
}

void SeaAnemonesApp::mouseDown(MouseEvent event)
{
    mCamera.mouseDown(event.getPos());
}

void SeaAnemonesApp::mouseDrag(MouseEvent event)
{
    mMousePos = event.getPos();
    mCamera.mouseDrag(event.getPos(), event.isLeftDown(), event.isMiddleDown(), event.isRightDown());
}

void SeaAnemonesApp::mouseUp( MouseEvent event )
{
}

void SeaAnemonesApp::update()
{
    if(mNormalMapCurrent != mNormalMapChosen){
        mNormalMapCurrent = mNormalMapChosen;
        mNormalMapTex = mNormalTextures[mNormalMapChosen]; //reference to a shared object, yay!
    }

    //updating time
    if(mAnimate) {
        //mTime = static_cast<float>(getElapsedSeconds() * 1000.0f);
        //float delta = mTime - mPrevTime;
        //mPrevTime = mTime;
        //mAnimTime += delta / 1000.f;

        mSeaAnemone.update((float)app::getElapsedSeconds());//mAnimTime);
    }

    if(mGradientFbo){
        mGradientFbo.bindFramebuffer();
        gl::clear();

        for(unsigned int i = 0; i < mSeaAnemone.colors.size() - 1; ++i){
            glBegin(GL_QUADS);
            glColor3f((Color&)mSeaAnemone.colors[i]);
            glVertex2f((GLfloat)getWindowWidth() * mSeaAnemone.colorStartsAt[i], 0);
            glVertex2f((GLfloat)getWindowWidth() * mSeaAnemone.colorStartsAt[i], (GLfloat)getWindowHeight());
            glColor3f((Color&)mSeaAnemone.colors[i+1]);
            glVertex2f((GLfloat)getWindowWidth() * mSeaAnemone.colorStartsAt[i+1], (GLfloat)getWindowHeight());
            glVertex2f((GLfloat)getWindowWidth() * mSeaAnemone.colorStartsAt[i+1], 0);
            glEnd();
        }

        mGradientFbo.unbindFramebuffer();
        mSeaAnemone.gradientColor = mGradientFbo.getTexture();
    }

    if(mSssOn){
        createDepthMap();
    }

    //zrobi to w ptli, adnie :)
    mLightPositions[0] = Vec3f(100,50,0);
    //mLightPositions[0] = Vec3f(180.f*cos(getElapsedSeconds()), 100.f*cos(getElapsedSeconds()*0.1f), 180.f*sin(getElapsedSeconds()));
    mLightPositions[1] = Vec3f(170.f*sin((float)getElapsedSeconds()), 50.f*cos((float)-getElapsedSeconds()*0.1f)+50.f, 150.f*cos((float)getElapsedSeconds()));
    mLightPositions[2] = Vec3f(160.f*sin((float)-getElapsedSeconds()), 75.f*cos((float)-getElapsedSeconds()*0.2f)+75.f, 160.f*sin((float)-getElapsedSeconds()));

    mLightColors[0] = Vec3f(0.3f, 0.7f, 0.2f);
    mLightColors[1] = Vec3f(0.f, 0.f, 1.f);
    mLightColors[2] = Vec3f(0.f, 1.f, 1.f);
}

void SeaAnemonesApp::draw()
{
    gl::clear();

    Area viewport = gl::getViewport();
    ((CameraPersp)mCamera.getCamera()).setAspectRatio(viewport.getWidth() / (float)viewport.getHeight());

    //gradient background
    glBegin(GL_QUADS);
    glColor3f(0.06f, 0.08f, 0.1f);
    glVertex2i(0, 0);
    glVertex2i(getWindowWidth(), 0);
    glColor3f(0.17f, 0.24f, 0.31f);
    glVertex2i(getWindowWidth(), getWindowHeight());
    glVertex2i(0, getWindowHeight());
    glEnd();

    if(mViewTexture)
        gl::draw(mTexture);

    gl::pushMatrices();
    gl::setMatrices(mCamera.getCamera());

    gl::enable(GL_CULL_FACE);

    gl::enableDepthRead();
    gl::enableDepthWrite();

    //apply material and setup light
    //change light later maybe
    gl::Light light(gl::Light::POINT, 0);
    light.setAmbient(Color::black());
    light.setDiffuse(Color::white());
    light.setSpecular(Color::white());
    light.setPosition(mLightPositions[0]);//mCamera.getCamera().getEyePoint());
    light.enable();

    for(int i=0; i<LIGHTS_COUNT; ++i){
        gl::color(Color(mLightColors[i].x, mLightColors[i].y, mLightColors[i].z));
        gl::drawSphere(mLightPositions[i], 10.f);
    }

    if(mDrawBloom){
        Area viewport = gl::getViewport();

        gl::setViewport(mFboScene.getBounds());
        mFboScene.bindFramebuffer();
        gl::pushMatrices();
        gl::setMatricesWindow(getWindowWidth(), getWindowHeight(), false);
        gl::clear(Color::black());
        renderAnemone(mCamera.getCamera().getEyePoint());
        gl::popMatrices();
        mFboScene.unbindFramebuffer();

        mShaderBlur.bind();
        mShaderBlur.uniform("tex0", 0);
        mShaderBlur.uniform("sampleOffset", Vec2f(1.0f/mFboBlur1.getWidth(), 0.0f));

        gl::setViewport( mFboBlur1.getBounds() );
        mFboBlur1.bindFramebuffer();
        mFboScene.bindTexture(0);
        gl::pushMatrices();
        gl::setMatricesWindow(64, 64, false);
        gl::clear( Color::black() );
        gl::drawSolidRect( mFboBlur1.getBounds() );
        gl::popMatrices();
        mFboScene.unbindTexture();
        mFboBlur1.unbindFramebuffer();
        mShaderBlur.uniform("sampleOffset", Vec2f(0.0f, 1.0f/mFboBlur2.getHeight()));

        gl::setViewport( mFboBlur2.getBounds() );
        mFboBlur2.bindFramebuffer();
        mFboBlur1.bindTexture(0);
        gl::pushMatrices();
        gl::setMatricesWindow(64, 64, false);
        gl::clear( Color::black() );
        gl::drawSolidRect( mFboBlur2.getBounds() );
        gl::popMatrices();
        mFboBlur1.unbindTexture();
        mFboBlur2.unbindFramebuffer();

        mShaderBlur.unbind();
        gl::setViewport(viewport);

        // because the Fbo's have their origin in the LOWER-left corner,
        // flip the Y-axis before drawing
        gl::pushModelView();
        gl::translate( Vec2f(0.f, (float)getWindowHeight()) );
        gl::scale( Vec3f(1.f, -1.f, 1.f) );

        gl::color(Color::white());
        gl::draw( mFboScene.getTexture(), Rectf(0.f, 0.f, (float)getWindowWidth(), (float)getWindowHeight()) );
        glEnable(GL_BLEND);
        glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR); //moe tutaj inczej flagi?
        gl::draw( mFboBlur2.getTexture(), Rectf(0.f, 0.f, (float)getWindowWidth(), (float)getWindowHeight()) );
        glDisable(GL_BLEND);

        gl::popModelView();
    }
    else{
        renderAnemone(mCamera.getCamera().getEyePoint());
    }

    mGroundMaterial.apply(); //czemu ta linia si nie wykonuje?
    gl::color(Color(0.2125f, 0.1275f, 0.054f));
    gl::drawCube(ci::Vec3f::zero(), ci::Vec3f(1000.0f, 1.f, 1000.0f));

    mSeaAnemone.drawDebug();

	gl::color(ci::Color::black());
	// draw all hair
	gl::drawSphere(mSeaAnemone.mBallPos,30,30);

	if(mDrawCapsules) {
		for(const auto &hair : mSeaAnemone.m_hairSet) {
			for(int i = 0; i < Hair::SEGMENT_COUNT; i++) {
				btTransform gt = hair.m_segments[i]->getWorldTransform();
				float glmat[16];
				gt.getOpenGLMatrix(glmat);
				ci::Matrix44f view(glmat);
		
				gl::pushModelView();
				gl::multModelView(view);

				gl::drawSphere(Vec3f(0,-hair.height/2.f,0),hair.radius);
				gl::drawSphere(Vec3f(0,hair.height/2.f,0),hair.radius);
				gl::translate(0,-hair.height/2.f,0);
				gl::drawCylinder(hair.radius,hair.radius,hair.height);
				gl::popModelView();
			}
		}
	}
    gl::disableDepthWrite();
    gl::disableDepthRead();
    gl::disable(GL_CULL_FACE);

    gl::popMatrices();

    //for testing only, covers whole screen
    if(mViewDepthTexture){
        gl::color(Color::white());
        gl::draw(mFboSceneAO.getTexture());
    }
    if(mViewNormalDepth){
        gl::draw(mFboSceneNormalDepth.getTexture());
    }

    //gl::drawString("FPS: " + to_string(getAverageFps()), Vec2f(5.0f, 5.0f), Color::white(), mFont);

    mParams->draw();
}

void SeaAnemonesApp::renderAnemone(Vec3f camera){
    if(mSssOn){
        if(mSssShader){
            mSssShader.bind();

            mSssShader.uniform("lightPosition", mLightPositions, LIGHTS_COUNT);
            mSssShader.uniform("lightColor", mLightColors, LIGHTS_COUNT);

            mSssShader.uniform("cameraPosition", camera);

            mSssShader.uniform("depthMap", 0);
            mSssShader.uniform("colorTex", 1);

            mSssShader.uniform("power", mPower);
            mSssShader.uniform("scale", mScale);
            mSssShader.uniform("distortion", mDisortion);
            mSssShader.uniform("ambient", mAmbient);
            mSssShader.uniform("shininess", mShininess);
            mSssShader.uniform("viewport", Vec2f(getWindowBounds().getSize()));
            mSssShader.uniform("diffuse", mDiffuse);

            mSssShader.uniform("kC", mKC);
            mSssShader.uniform("kL", mKL);
            mSssShader.uniform("kQ", mKQ);
            mSssShader.uniform("specScale", mSpecScale);

            glEnableVertexAttribArray(attr_height);
            glVertexAttribPointer(attr_height, 1, GL_FLOAT, GL_FALSE, sizeof(GLfloat), &mSeaAnemone.heights[0]);

            mFboSceneAO.getTexture().bind(0);
            mSeaAnemone.gradientColor.bind(1);
			
			//mSeaAnemone.draw();
            mSeaAnemone.drawBasalDisc();
            glVertexAttribPointer(attr_height, 1, GL_FLOAT, GL_FALSE, sizeof(GLfloat), &mSeaAnemone.heightsHair[0]);
			mSeaAnemone.drawCnidocytes();

            mFboSceneAO.getTexture().unbind();
            mSeaAnemone.gradientColor.unbind();
            mSssShader.unbind();
            glDisableVertexAttribArray(attr_height);
        }
        else
            renderAnemoneGradient();
    }
    else
        renderAnemoneGradient();
}

void SeaAnemonesApp::renderAnemoneGradient(){
    // bind shader
    if(mBumpMapOn){
        if(mBumpGradientShader){
            mBumpGradientShader.bind();

            //these should be moved to update function
            mBumpGradientShader.uniform("tex0", 0);
            mBumpGradientShader.uniform("tex1", 1);
            mBumpGradientShader.uniform("tex2", 2);

            mBumpGradientShader.uniform("texAlpha", mTextureAlpha);
            glEnableVertexAttribArray(attr_height);
            glVertexAttribPointer(attr_height, 1, GL_FLOAT, GL_FALSE, sizeof(GLfloat), &mSeaAnemone.heights[0]);
        }

        mTexture.bind(0);
        mSeaAnemone.gradientColor.bind(1);
        mNormalMapTex.bind(2);

        mMaterial.apply();
        			//mSeaAnemone.draw();
            mSeaAnemone.drawBasalDisc();
            glVertexAttribPointer(attr_height, 1, GL_FLOAT, GL_FALSE, sizeof(GLfloat), &mSeaAnemone.heightsHair[0]);
			mSeaAnemone.drawCnidocytes();

        mTexture.unbind();
        mSeaAnemone.gradientColor.unbind();
        mNormalMapTex.unbind();

        // unbind shader
        if(mBumpGradientShader){
            mBumpGradientShader.unbind();
            glDisableVertexAttribArray(attr_height);
        }
    }
    else{
        if(mPhongShader){
            mPhongShader.bind();

            //these should be moved to update function
            mPhongShader.uniform("tex0", 0);
            mPhongShader.uniform("tex1", 1);

            mPhongShader.uniform("texAlpha", mTextureAlpha);
            mPhongShader.uniform("numLights", LIGHTS_COUNT);

            glEnableVertexAttribArray(attr_height);
            glVertexAttribPointer(attr_height, 1, GL_FLOAT, GL_FALSE, sizeof(GLfloat), &mSeaAnemone.heights[0]);
        }

        mTexture.bind(0);
        mSeaAnemone.gradientColor.bind(1);

        mMaterial.apply();
                			//mSeaAnemone.draw();
            mSeaAnemone.drawBasalDisc();
            glVertexAttribPointer(attr_height, 1, GL_FLOAT, GL_FALSE, sizeof(GLfloat), &mSeaAnemone.heightsHair[0]);
			mSeaAnemone.drawCnidocytes();

        mTexture.unbind();
        mSeaAnemone.gradientColor.unbind();

        // unbind shader
        if(mPhongShader){
            mPhongShader.unbind();
            glDisableVertexAttribArray(attr_height);
        }
    }
}

void SeaAnemonesApp::createDepthMap(){
    if(mNormalDepthShader){
        gl::setViewport(mFboSceneNormalDepth.getBounds());
        gl::enable(GL_CULL_FACE);
        gl::enableDepthRead();
        gl::enableDepthWrite();

        mFboSceneNormalDepth.bindFramebuffer();
        gl::pushMatrices();
        gl::setMatrices(mCamera.getCamera());

        gl::clear();
        gl::color(Color::white());

        mNormalDepthShader.bind();
        mNormalDepthShader.uniform("farPlane", farPlane);

                			//mSeaAnemone.draw();
            mSeaAnemone.drawBasalDisc();
            glVertexAttribPointer(attr_height, 1, GL_FLOAT, GL_FALSE, sizeof(GLfloat), &mSeaAnemone.heightsHair[0]);
			mSeaAnemone.drawCnidocytes();

        mNormalDepthShader.unbind();

        gl::popMatrices();
        mFboSceneNormalDepth.unbindFramebuffer();

        gl::disableDepthWrite();
        gl::disableDepthRead();
        gl::disable(GL_CULL_FACE);

        gl::setViewport(getWindowBounds());
    }

    if(mSSAOShader){
        Area viewport = gl::getViewport();

        gl::setViewport(mFboSceneAO.getBounds());
        gl::enable(GL_CULL_FACE);
        gl::enableDepthRead();
        gl::enableDepthWrite();

        mFboSceneAO.bindFramebuffer();
        gl::setMatricesWindow(mFboSceneAO.getSize());

        gl::clear();
        gl::color(Color::white());

        mSSAOShader.bind();

        mFboSceneNormalDepth.getTexture().bind(0);
        mNoiseTextures[0].bind(1);
        mSSAOShader.uniform("normalMap", 0);
        mSSAOShader.uniform("rnm", 1);
        mSSAOShader.uniform("finalStrength", mFinalStrength);
        mSSAOShader.uniform("totStrength", mTotStrength);
        mSSAOShader.uniform("strength", mStrength);
        mSSAOShader.uniform("offset", mOffset);
        mSSAOShader.uniform("falloff", mFalloff);
        mSSAOShader.uniform("rad", mRad);

        //mSeaAnemone.draw();
        gl::drawSolidRect(Rectf(0.f, 0.f, (float)getWindowWidth(), (float)getWindowHeight()));

        mSSAOShader.unbind();
        mFboSceneNormalDepth.getTexture().unbind();
        mNoiseTextures[0].unbind();

        gl::popMatrices();

        mFboSceneAO.unbindFramebuffer();

        gl::disableDepthWrite();
        gl::disableDepthRead();
        gl::disable(GL_CULL_FACE);
        gl::setViewport(getWindowBounds());
    }
}

void SeaAnemonesApp::renderAnemoneBump(){
    if(mBumpMapOn){
        if(mBumpShader){
            mBumpShader.bind();

            //these should be moved to update function
            mBumpShader.uniform("Texture", 0);

            //mBumpShader.uniform("texAlpha", mTextureAlpha);
        }

        mNormalMapTex.bind(0);

        //mMaterial.apply();
                			//mSeaAnemone.draw();
            mSeaAnemone.drawBasalDisc();
            glVertexAttribPointer(attr_height, 1, GL_FLOAT, GL_FALSE, sizeof(GLfloat), &mSeaAnemone.heightsHair[0]);
			mSeaAnemone.drawCnidocytes();

        mNormalMapTex.unbind();

        mBumpShader.unbind();
    }
}

CINDER_APP_NATIVE(SeaAnemonesApp, RendererGl)