#include "PrecompiledHeaders.h"
#include "ShapeGenerator.h"

ShapeGenerator::ShapeGenerator(std::vector<Vec3f>& verticesRef, std::vector<Vec3f>& normalsRef, std::vector<uint32_t>& trianglesRef, 
							   std::vector<Vec2f>& texcoordsRef, std::vector<float>& heightsRef)
    : vertices(verticesRef), normals(normalsRef), triangles(trianglesRef), texCoords(texcoordsRef), heights(heightsRef){
}

void ShapeGenerator::buildCylinder(Vec3f center, float H, float radius, int radialSegments, int heightSegments, 
                                   Quaternion<float> rotation, int numWaves, float waveD){
    float dheight = H / heightSegments;

    for (int i = 0; i <= heightSegments; i++)
    {
        Vec3f centrePos = center + Vec3f::yAxis() * dheight * i;
        float v = (float)i / heightSegments;
        buildCylinderRing(radialSegments, centrePos, radius, rotation, v, i > 0, false, numWaves, waveD);
        if (i==0)
            buildCap(radialSegments, centrePos, radius, rotation, true, false, numWaves, waveD);
        else if (i==heightSegments)
            buildCap(radialSegments, centrePos, radius, rotation, false, false, numWaves, waveD);
    }
}

void ShapeGenerator::buildCylinderRing(int segments, Vec3f center, float radius, Quaternion<float> rotation, float v, bool buildTriangles, 
                                       bool waves, int numWaves, float waveD, float wavesH, float height){
    float dangle = (float)((M_PI * 2.0f) / segments);
    for (int i = 0; i <= segments; i++)
    {
        float angle = dangle * i;
        Vec3f position = Vec3f::zero();
        position.x = cos(angle);
        position.z = sin(angle);

        if (waves){
            position.x *= (float)(1 + waveD * cos((M_PI-angle) * numWaves) * wavesH);
            position.z *= (float)(1 + waveD * cos((M_PI-angle) * numWaves) * wavesH);
        }
		if(height<0) {
			heights.push_back(position.y + center.y);
		} else {
			heights.push_back(height);
		}

        position = rotation * position;

        Vec3f vec = center + position * radius;

        vertices.push_back(vec);
        normals.push_back(position);
        texCoords.push_back(Vec2f((float)i / segments, v));

        if (i > 0 && buildTriangles)
        {
            int baseIndex = vertices.size()-1;
            int vertsPerRow = segments + 1;

            int index0 = baseIndex;
            int index1 = baseIndex - 1;
            int index2 = baseIndex - vertsPerRow;
            int index3 = baseIndex - vertsPerRow - 1;
            addTriangle(index0, index2, index1);
            addTriangle(index2, index3, index1);
        }
    }
}

void ShapeGenerator::buildCap(int segments, Vec3f center, float radius, Quaternion<float> rotation, bool reverse, 
                              bool waves, int numWaves, float waveD, float wavesH, float height){
    Vec3f normal = reverse ? -Vec3f::yAxis() : Vec3f::yAxis();

    vertices.push_back(center);
    normals.push_back(normal);
    texCoords.push_back(Vec2f(0.5f, 0.5f));
	if(height<0) {
		heights.push_back(center.y);
	} else {
		heights.push_back(height);
	}

    int centreVertexIndex = vertices.size()-1;

    float dangle = (float)((M_PI * 2.0f) / segments);
    for (int i = 0; i <= segments; i++)
    {
        float angle = dangle * i;

        Vec3f position = Vec3f::zero();
        position.x = cos(angle);
        position.z = sin(angle);

        if (waves){
            position.x *= (float)(1 + waveD * cos((M_PI-angle) * numWaves) * wavesH);
            position.z *= (float)(1 + waveD * cos((M_PI-angle) * numWaves) * wavesH);
        }
		if(height<0) {
			heights.push_back(position.y + center.y);
		} else {
			heights.push_back(height);
		}

        position = rotation * position;

        vertices.push_back(center + position * radius);
        normals.push_back(normal);

        Vec2f uv = Vec2f(position.x + 1.0f, position.z + 1.0f) * 0.5f;
        texCoords.push_back(uv);

        if (i > 0)
        {
            int baseIndex = vertices.size()-1;

            if (reverse)
                addTriangle(centreVertexIndex, baseIndex - 1, baseIndex);
            else
                addTriangle(centreVertexIndex, baseIndex, baseIndex - 1);
        }
    }
}

void ShapeGenerator::buildCone(Vec3f center, float H, float R_start, float R_end, int radialSegments, int heightSegments, Quaternion<float> rotation, Vec3f h_axis = Vec3f::yAxis()){    float dheight = H / heightSegments;

    Vec2f slope = Vec2f(R_end - R_start, H);
    slope.normalize();

    for (int i = 0; i <= heightSegments; i++)
    {
        Vec3f centrePos = center + h_axis * dheight * i;
        float radius = lerp(R_start, R_end, (float)i / heightSegments);

        float v = (float)i / heightSegments;
        buildConeRing(radialSegments, centrePos, radius, rotation, v, i > 0, slope);
    }
}

void ShapeGenerator::buildConeRing(int segments, Vec3f center, float radius, Quaternion<float> rotation, float v, bool buildTriangles, Vec2f slope, float height){
    float dangle = (float)((M_PI * 2.0f) / segments);
    for (int i = 0; i <= segments; i++)
    {
        float angle = dangle * i;
        Vec3f position = Vec3f::zero();
        position.x = cos(angle);
        position.z = sin(angle);

        float normalVertical = -slope.x;
		float normalHorizontal = slope.y;

		Vec3f normal = position * normalHorizontal;
		normal.y = normalVertical;
		normal = rotation * normal;

		if(height < 0) {
			heights.push_back(position.y + center.y);
		} else {
			heights.push_back(height);
		}
        position = rotation * position;

        Vec3f vec = center + position * radius;

        vertices.push_back(vec);
        normals.push_back(normal);
        texCoords.push_back(Vec2f((float)i / segments, v));
        if (i > 0 && buildTriangles)
        {
            int baseIndex = vertices.size()-1;
            int vertsPerRow = segments + 1;

            int index0 = baseIndex;
            int index1 = baseIndex - 1;
            int index2 = baseIndex - vertsPerRow;
            int index3 = baseIndex - vertsPerRow - 1;
            addTriangle(index0, index2, index1);
            addTriangle(index2, index3, index1);
        }
    }
}

void ShapeGenerator::buildFlatSphere(Vec3f center, float radius, Quaternion<float> rotation, int radialSegments, int heightSegments, float verticalScale, float height){
    float angleInc = (float)(M_PI / heightSegments);
	float verticalRadius = radius * verticalScale;
    //float hairLine = hairOnTop * heightSegments;

	for (int i = heightSegments/2; i <= heightSegments; i++)
	{
		Vec3f centrePos = Vec3f::zero();

		centrePos.y = -cos(angleInc * i);
		float rad= sin(angleInc * i);

		Vec2f slope = Vec2f(-centrePos.y / verticalScale, rad);
		slope.normalize();

		centrePos.y = centrePos.y * verticalRadius + verticalRadius;
		rad *= radius;

		Vec3f finalRingCentre = rotation * centrePos + center;
		float v = (float)i / heightSegments;

        buildConeRing(radialSegments, finalRingCentre, rad, rotation, v, i > 0, slope, height);
        //if(i>=hairLine){
        //    buildHair(finalRingCentre, rotation, radius);
        //}
    }
}

void ShapeGenerator::buildSphere(Vec3f centerBottom, float radius, int radialSegments, int heightSegments, float height1){
    Quaternion<float> rotation = Quaternion<float>();

    int heightSegmentCount = radialSegments / 2;
    float angleInc = (float)(M_PI / heightSegments);

    for (int i = 0; i <= heightSegments; i++)
    {
        Vec3f centrePos = centerBottom;

        centrePos.y += -cos(angleInc * i) * radius;
        float rad = sin(angleInc * i) * radius;

        float v = (float)i / heightSegmentCount;
        buildSphereRing(radialSegments, centrePos, rad, rotation, v, i>0, height1);
    }
}

void ShapeGenerator::buildSphereRing(int segments, Vec3f center, float radius, Quaternion<float> rotation, float v, bool buildTriangles, float height){
    float dangle = (float)((M_PI * 2.0f) / segments);
    for (int i = 0; i <= segments; i++)
    {
        float angle = dangle * i;
        Vec3f position = Vec3f::zero();
        position.x = cos(angle);
        position.z = sin(angle);

        Vec3f vec = center + position * radius;
		if(height < 0) {
			heights.push_back(vec.y);
		} else {
			heights.push_back(height);
		}

        vec = rotation * vec;
        //todo: move back when rotating

        vertices.push_back(vec);
        normals.push_back(vec.normalized());
        texCoords.push_back(Vec2f((float)i / segments, v));

        if (i > 0 && buildTriangles)
        {
            int baseIndex = vertices.size()-1;
            int vertsPerRow = segments + 1;

            int index0 = baseIndex;
            int index1 = baseIndex - 1;
            int index2 = baseIndex - vertsPerRow;
            int index3 = baseIndex - vertsPerRow - 1;
            addTriangle(index0, index2, index1);
            addTriangle(index2, index3, index1);
        }
    }
}

void ShapeGenerator::addTriangle(int a, int b, int c){
    triangles.push_back(a);
    triangles.push_back(b);
    triangles.push_back(c);
}