#include "interpolation.h"

namespace modGeo {

interpolation::interpolation(void) : mLineLengthSum(0.0) { }

interpolation::~interpolation(void) { }

void interpolation::addInterPoint(vector inPoint) {
	mInterPoints.push_back(inPoint);
}

void interpolation::moveInterPoint(int in_n, vector inPoint) {
	assert(in_n >= 0 && in_n < (int)mInterPoints.size());
	mInterPoints[in_n] = inPoint;
}

void interpolation::removeLastInterPoint() {
	assert((int)mInterPoints.size() > 0);
	mInterPoints.pop_back();
}

ci::Vec3f interpolation::deCasteljauUniform(double t) const {
	t *= mLineLengthSum;
		
	double current_t = mLineLengths[0];
	double sum_t = 0.0;
	int i = 0;
	while(t > current_t) {
		sum_t += mLineLengths[i];
		current_t += mLineLengths[++i];
	}

	t = (t - sum_t) / mLineLengths[i];
	vector v = deCasteljau(i,t);
	return ci::Vec3f((float)v[0],(float)v[1],(float)v[2]);
}

void interpolation::updateControlPoints() {
	double lTotalDistance = 0.0;
	const int n = mInterPoints.size();
	if (n <= 1) return;

	std::vector<double> lKnots(n);
	mLineLengths.clear();

	lKnots[0] = 0; // at least one interpolation point
	for(int j = 1; j < n; ++j) {
		double lDistance = distance(mInterPoints[j], mInterPoints[j-1]);
		lDistance = std::max(lDistance, 0.01); // TODO: division by 0 fix. workaround now.
		lTotalDistance += lDistance;
		lKnots[j] = lTotalDistance;

		mLineLengths.push_back(lDistance);
	}
	mLineLengthSum = lTotalDistance;

	double lDistMultiplier = 100.0 / lTotalDistance;
	for(int j = 1; j < (int)mInterPoints.size(); ++j) {
		lKnots[j] *= lDistMultiplier;
	}

	std::vector<vector> b(mInterPoints.size());
	std::vector<vector>& y = mInterPoints;
	std::vector<double>& x = lKnots;
	for(int i = 1; i < n-1; ++i) {
		b[i] = ((y[i] - y[i-1]) / pow(x[i] - x[i-1], 2.0) + (y[i+1] - y[i]) / pow(x[i+1] - x[i], 2.0)) * 3.0;
	}
	b[0] = (y[1] - y[0]) / pow(x[1] - x[0], 2.0) * 3.0;
	b[n-1] = (y[n-1] - y[n-2]) / pow(x[n-1] - x[n-2], 2.0) * 3.0;

	std::vector<vector> diag(n);
	std::vector<vector> ldiag(n);
	std::vector<vector> rdiag(n);
	const vector vone(1.0, 1.0, 1.0);
	for(int i = 1; i < n-1; ++i) {
		ldiag[i] = vone / (x[i] - x[i-1]);
		diag[i] = (vone / (x[i] - x[i-1]) + vone / (x[i+1] - x[i])) * 2.0;
		rdiag[i] = vone / (x[i+1] - x[i]);
	}
	diag[0] = vone / (x[1] - x[0]) * 2;
	rdiag[0] = vone / (x[1] - x[0]);
	ldiag[n-1] = vone / (x[n-1] - x[n-2]);
	diag[n-1] = vone / (x[n-1] - x[n-2]) * 2;

	//
	b[1] = b[1] - mFirstTangent * ldiag[1];
	//

	std::vector<vector> v = b;
	rdiag[1] = rdiag[1] / diag[1];
	v[1] = v[1] / diag[1];
	for(int i = 2; i < n; i++) {
		vector m = vone / (diag[i] - ldiag[i] * rdiag[i-1]);
		rdiag[i] = rdiag[i] * m;
		v[i] = (v[i] - ldiag[i] * v[i-1]) * m;
	}
	for(int i = n-1; i-- > 1; ) {
		v[i] = v[i] - rdiag[i] * v[i+1];
	}

	std::vector<vector> bezierPoints;
	for(int i = 1; i < n; i++) {
		vector a = v[i-1] * (x[i] - x[i-1]) - (y[i] - y[i-1]);
		vector b =  v[i] * -(x[i] - x[i-1]) + (y[i] - y[i-1]);
		// polynomial base
		vector t0 = y[i-1];
		vector t1 = y[i] - y[i-1] + a;
		vector t2 = b - a * 2;
		vector t3 = a - b;
		// to bernstein base
		vector aa = t0;
		vector bb = t1 / 3.0 + aa;
		vector cc = t2 / 3.0 - aa + bb * 2.0;
		vector dd = t3 + aa - bb * 3.0 + cc * 3.0;
		bezierPoints.push_back(aa);
		bezierPoints.push_back(bb);
		bezierPoints.push_back(cc);
		bezierPoints.push_back(dd);
	}
	mBezierPoints = bezierPoints;
}

int interpolation::getPointsLength() const {
	return (int)mInterPoints.size();
}

vector interpolation::deCasteljau(int i, double t) const {
	const double u = 1 - t;
	const int n = 3;
	std::vector<vector> b(n+1);
	std::copy(mBezierPoints.begin() + i * (n + 1), 
		mBezierPoints.begin() + i * (n + 1) + n + 1, 
		begin(b));
	for(int i = 0; i < n; ++i) {
		for(int j = 0; j < n - i; ++j) {
			b[j] = b[j] * u + b[j+1] * t;
		}
	}
	return b[0];
}

void interpolation::addInterPoint(ci::Vec3f point) {
	addInterPoint(vector(point.x, point.y, point.z));
}

void interpolation::moveInterPoint(int n, ci::Vec3f point) {
	moveInterPoint(n, vector(point.x, point.y, point.z));
}
void interpolation::setFirstTangent(ci::Vec3f Tangent) {
	mFirstTangent = vector(Tangent.x, Tangent.y, Tangent.z);
}

ci::Vec3f interpolation::tangent(double t) const {
	const double h = 0.00006103515;
	const double h2 = 2.0 * h;
	ci::Vec3f v(1,0,0);
	if(t-h < 0.0) {
		v = (deCasteljauUniform(t+h) - deCasteljauUniform(t)) / (float)h;
	} else if(t+h > 1.0) {
		v = (deCasteljauUniform(t) - deCasteljauUniform(t-h)) / (float)h;
	} else {
		v = (deCasteljauUniform(t+h) - deCasteljauUniform(t-h)) / (float)h2;
	}
	return v.normalized();
}

ci::Vec3f interpolation::normal(double t) const {
	const double h = 0.00006103515;
	const double h2 = 2.0 * h;
	if(t-h < 0.0) {
		return ((tangent(t+h) - tangent(t)) / (float)h).normalized();
	} else if(t-h > 1.0) {
		return ((tangent(t) - tangent(t-h)) / (float)h).normalized();
	} else {
		return ((tangent(t+h) - tangent(t-h)) / (float)h2).normalized();
	}
}

ci::Vec3f interpolation::binormal(double t) const {
	return ci::cross(tangent(t),normal(t));
}

}
